/*
 * Copyright 2012 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// Author: piatek@google.com (Michael Piatek)

#ifndef NET_INSTAWEB_HTTP_PUBLIC_REQUEST_CONTEXT_H_
#define NET_INSTAWEB_HTTP_PUBLIC_REQUEST_CONTEXT_H_

#include "net/instaweb/util/public/basictypes.h"
#include "net/instaweb/util/public/ref_counted_ptr.h"
#include "net/instaweb/util/public/scoped_ptr.h"
#include "net/instaweb/util/public/string_util.h"

namespace net_instaweb {

class AbstractLogRecord;
class AbstractMutex;
class RequestContext;
class RequestTrace;
class ThreadSystem;
class Timer;

typedef RefCountedPtr<RequestContext> RequestContextPtr;

// A class which wraps state associated with a request.
//
// This object should be reference counted, wrapped in a RequestContextPtr. We
// use reference counting because, depending on the timing of asynchronous
// rewrites, RPC calls, and so on, a RequestContext may outlive the original
// HTTP request serving, or not. Reference counting avoids the complexity of
// explicit transfer of ownership in these cases.
class RequestContext : public RefCounted<RequestContext> {
 public:
  // |logging_mutex| will be passed to the request context's AbstractLogRecord,
  // which will take ownership of it. If you will be doing logging in a real
  // (threaded) environment, pass in a real mutex. If not, a NullMutex is fine.
  // |timer| will be passed to the TimingInfo, which will *not* take ownership.
  // Passing NULL for |timer| is allowed.
  explicit RequestContext(AbstractMutex* logging_mutex, Timer* timer);

  // TODO(marq): Move this test context factory to a test-specific file.
  //             Makes a request context for running tests.
  static RequestContextPtr NewTestRequestContext(ThreadSystem* thread_system);

  // Creates a new, unowned AbstractLogRecord, for use by some subordinate
  // action.  Also useful in case of background activity where logging is
  // required after the response is written out, e.g., blink flow.
  virtual AbstractLogRecord* NewSubordinateLogRecord(
      AbstractMutex* logging_mutex);

  // The root trace context is associated with the user request which we
  // are attempting to serve. If this is a request with constituent resources
  // that we rewrite, there may be several dependent fetches synthesized
  // by PSOL during rewrites. Those are traced separately.
  RequestTrace* root_trace_context() { return root_trace_context_.get(); }
  // Takes ownership of the given context.
  void set_root_trace_context(RequestTrace* x);

  // Creates a new RequestTrace associated with a request depending on the
  // root user request; e.g., a subresource fetch for an HTML page.
  //
  // This implementation is a no-op. Subclasses should customize this based
  // on their underlying tracing system. A few interface notes:
  // - The caller is not responsible for releasing memory or managing the
  //   lifecycle of the RequestTrace.
  // - A call to CreateDependentTraceContext() need not be matched by a call
  //   to ReleaseDependentTraceContext(). Cleanup should be automatic and
  //   managed by RequestContext subclass implementations.
  virtual RequestTrace* CreateDependentTraceContext(const StringPiece& label) {
    return NULL;
  }

  // Releases this object's reference to the given context and frees memory.
  // Calls to CreateDependentTraceContext need not be matched by
  // calls to this function. If a dependent trace span is not released when
  // the request context reference count drops to zero, this object will clean
  // all dependent traces.
  //
  // Note that automatic cleanup of dependent traces is provided for safety.
  // To provide meaningful performance statistics, cleanup should be
  // coupled with the completion of the event being traced.
  //
  // Subclasses should customize this based on their underlying tracing system.
  virtual void ReleaseDependentTraceContext(RequestTrace* t);

  // The log record for the this request, created when the request context is.
  virtual AbstractLogRecord* log_record();

  // Determines whether this request is using the SPDY protocol.
  bool using_spdy() const { return using_spdy_; }
  void set_using_spdy(bool x) { using_spdy_ = x; }

  // Prepare the AbstractLogRecord for a subsequent call to WriteLog.  This
  // might include propagating information collected in the RequestContext,
  // TimingInfo for example, to the underlying logging infrastructure.
  void PrepareLogRecordForOutput();

  // Write the log for background rewriting into disk.
  void WriteBackgroundRewriteLog();

  // Return the log record for background rewrites. If it doesn't exist, create
  // a new one.
  AbstractLogRecord* GetBackgroundRewriteLog(
      ThreadSystem* thread_system,
      bool log_urls,
      bool log_url_indices,
      int max_rewrite_info_log_size);

  // TODO(gee): Track queuing time.
  class TimingInfo {
   public:
    // TODO(gee): Does this need to be thread safe?
    // Timer is not owned by TimingInfo.
    TimingInfo();

    // Initialize the TimingInfo with the specified Timer.  Sets a timestamp
    // from which GetElapsedMs is based.
    // TODO(gee): This is forced by the RequestContext/GoogleRequestContext
    void Init(Timer* timer);

    // This should be called when the request "starts", for example in
    // HTTPServerFetch::Start.  It denotes the request "start time", which all
    // other timing values are relative to.
    void RequestStarted();

    // Called when the request is finished, i.e. the response has been sent to
    // the client.
    void RequestFinished();

    // Fetch related timing events.
    // Note:  Only the first call to FetchStarted will have an affect,
    // subsequent calls are silent no-ops.
    void FetchStarted();
    void FetchFirstByteReceived();
    void FetchHeaderReceived();
    void FetchFinished();

    // Milliseconds since Init.
    int64 GetElapsedMs() const;

    // Milliseconds since FetchStarted.
    int64 GetElapsedFromFetchStart();

    // Methods for retrieving information.
    int64 init_ts_ms() const {
      return init_ts_ms_;
    }

    int64 start_ts_ms() const {
      return start_ts_ms_;
    }

    // Milliseconds from request start to fetch start.
    int64 fetch_start_ms() const { return fetch_start_ts_ms_ - start_ts_ms_; }

    int64 fetch_first_byte_ms() const { return fetch_first_byte_ms_; }

    // Milliseconds from fetch start to head received.
    int64 fetch_header_ms() const { return fetch_header_ms_; }

    // Milliseconds from fetch start to fetch end.
    int64 fetch_elapsed_ms() const { return fetch_elapsed_ms_; }

    // Milliseconds spent "processing": end time - start time - fetch time.
    int64 processing_elapsed_ms() const { return processing_elapsed_ms_; }

   private:
    int64 NowMs() const;

    Timer* timer_;

    // Timestamps.
    int64 init_ts_ms_;

    int64 start_ts_ms_;

    int64 fetch_start_ts_ms_;

    // Durations.
    // TODO(gee): Make all members timestamps?
    int64 fetch_first_byte_ms_;
    int64 fetch_header_ms_;
    int64 fetch_elapsed_ms_;

    int64 processing_elapsed_ms_;

    DISALLOW_COPY_AND_ASSIGN(TimingInfo);
  };

  const TimingInfo& timing_info() const { return timing_info_; }
  TimingInfo* mutable_timing_info() { return &timing_info_; }

 protected:
  // TODO(gee): Fix this, it sucks.
  // The default constructor will not create a LogRecord. Subclass constructors
  // must do this explicitly.
  RequestContext();
  // Destructors in refcounted classes should be protected.
  virtual ~RequestContext();
  REFCOUNT_FRIEND_DECLARATION(RequestContext);

 private:
  // Always non-NULL.
  scoped_ptr<AbstractLogRecord> log_record_;

  TimingInfo timing_info_;

  // Logs tracing events associated with the root request.
  scoped_ptr<RequestTrace> root_trace_context_;

  // Log for recording background rewritings.
  scoped_ptr<AbstractLogRecord> background_rewrite_log_record_;

  bool using_spdy_;

  DISALLOW_COPY_AND_ASSIGN(RequestContext);
};

}  // namespace net_instaweb

#endif  // NET_INSTAWEB_HTTP_PUBLIC_REQUEST_CONTEXT_H_
