/*
 * Copyright 2013 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// Author: pulkitg@google.com (Pulkit Goyal)
//
// This class holds property page with fallback values (i.e. without query
// params) which can be used to retrieve and update the property values in case
// if we fail to find it in actual property page. Values retrieved from property
// page without query params are named as fallback values.

#ifndef NET_INSTAWEB_UTIL_PUBLIC_FALLBACK_PROPERTY_PAGE_H_
#define NET_INSTAWEB_UTIL_PUBLIC_FALLBACK_PROPERTY_PAGE_H_

#include "net/instaweb/util/public/basictypes.h"
#include "net/instaweb/util/public/scoped_ptr.h"
#include "net/instaweb/util/public/cache_interface.h"
#include "net/instaweb/util/public/property_cache.h"
#include "net/instaweb/util/public/string.h"
#include "net/instaweb/util/public/string_util.h"

namespace net_instaweb {

class FallbackPropertyPage : public AbstractPropertyPage {
 public:
  // FallbackPropertyPage takes the ownership of both the property pages.
  FallbackPropertyPage(PropertyPage* actual_property_page,
                       PropertyPage* property_page_with_fallback_values);
  virtual ~FallbackPropertyPage();

  // Gets a property given the property name. It returns the property
  // from fallback property cache if actual property page has no value.
  virtual PropertyValue* GetProperty(
      const PropertyCache::Cohort* cohort,
      const StringPiece& property_name) const;

  // Updates the value of a property for both actual property page and fallback
  // property page.
  virtual void UpdateValue(
      const PropertyCache::Cohort* cohort, const StringPiece& property_name,
      const StringPiece& value);

  // Updates a Cohort of properties into the cache. It will also update for
  // fallback property cache.
  virtual void WriteCohort(const PropertyCache::Cohort* cohort);

  // Gets the cache state for the actual property page.
  virtual CacheInterface::KeyState GetCacheState(
      const PropertyCache::Cohort* cohort);

  // Deletes a property given the property name from both the pages.
  virtual void DeleteProperty(const PropertyCache::Cohort* cohort,
                              const StringPiece& property_name);

  virtual const GoogleString& key() const;

  PropertyPage* actual_property_page() { return actual_property_page_.get(); }
  PropertyPage* property_page_with_fallback_values() {
    return property_page_with_fallback_values_.get();
  }

 private:
  scoped_ptr<PropertyPage> actual_property_page_;
  scoped_ptr<PropertyPage> property_page_with_fallback_values_;
  DISALLOW_COPY_AND_ASSIGN(FallbackPropertyPage);
};

}  // namespace net_instaweb
#endif  // NET_INSTAWEB_UTIL_PUBLIC_FALLBACK_PROPERTY_PAGE_H_
